@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.026';
+our $VERSION = '0.027';
 
 use EV;
 use Time::HiRes;
@@ -20,6 +20,8 @@ sub YAHC::Error::NO_ERROR                () { 0 }
 sub YAHC::Error::REQUEST_TIMEOUT         () { 1 << 0 }
 sub YAHC::Error::CONNECT_TIMEOUT         () { 1 << 1 }
 sub YAHC::Error::DRAIN_TIMEOUT           () { 1 << 2 }
+sub YAHC::Error::LIFETIME_TIMEOUT        () { 1 << 3 }
+sub YAHC::Error::RETRY_LIMIT             () { 1 << 9 }
 
 sub YAHC::Error::CONNECT_ERROR           () { 1 << 10 }
 sub YAHC::Error::READ_ERROR              () { 1 << 11 }
@@ -28,11 +30,12 @@ sub YAHC::Error::REQUEST_ERROR           () { 1 << 13 }
 sub YAHC::Error::RESPONSE_ERROR          () { 1 << 14 }
 sub YAHC::Error::CALLBACK_ERROR          () { 1 << 15 }
 sub YAHC::Error::SSL_ERROR               () { 1 << 16 }
+sub YAHC::Error::TERMINAL_ERROR          () { 1 << 30 }
 sub YAHC::Error::INTERNAL_ERROR          () { 1 << 31 }
 
 sub YAHC::State::INITIALIZED             () { 0   }
 sub YAHC::State::RESOLVE_DNS             () { 5   }
-sub YAHC::State::WAIT_SYNACK             () { 10  }
+sub YAHC::State::CONNECTING              () { 10  }
 sub YAHC::State::CONNECTED               () { 15  }
 sub YAHC::State::SSL_HANDSHAKE           () { 20  }
 sub YAHC::State::WRITING                 () { 25  }
@@ -48,13 +51,14 @@ use constant {
     # data from only a single SSL frame you can guarantee that there are no
     # pending data.
     TCP_READ_CHUNK              => 131072,
-    CALLBACKS                   => [ qw/init_callback wait_synack_callback connected_callback
+    CALLBACKS                   => [ qw/init_callback connecting_callback connected_callback
                                         writing_callback reading_callback callback/ ],
 };
 
 our @EXPORT_OK = qw/
     yahc_retry_conn
     yahc_reinit_conn
+    yahc_terminal_error
     yahc_conn_last_error
     yahc_conn_id
     yahc_conn_url
@@ -82,7 +86,8 @@ sub new {
 
     # wrapping target selection here allows all client share same list
     # and more importantly to share index within the list
-    $args->{_target} = _wrap_target_selection($args->{host}) if $args->{host};
+    $args->{_target}  = _wrap_host($args->{host})             if $args->{host};
+    $args->{_backoff} = _wrap_backoff($args->{backoff_delay}) if $args->{backoff_delay};
 
     my %storage;
     my $self = bless {
@@ -123,13 +128,19 @@ sub request {
     do { $request->{$_} ||= $pool_args->{$_} if $pool_args->{$_} } foreach (qw/host port scheme request_timeout
                                                                                connect_timeout drain_timeout/);
     if ($request->{host}) {
-        $request->{_target} = _wrap_target_selection($request->{host});
+        $request->{_target} = _wrap_host($request->{host});
     } elsif ($pool_args->{_target}) {
         $request->{_target} = $pool_args->{_target};
     } else {
         die "YAHC: host must be defined in request() or in new()\n";
     }
 
+    if ($request->{backoff_delay}) {
+        $request->{_backoff} = _wrap_backoff($request->{backoff_delay});
+    } elsif ($pool_args->{_backoff}) {
+        $request->{_backoff} = $pool_args->{_backoff};
+    }
+
     my $scheme = $request->{scheme} ||= 'http';
     my $debug = delete $request->{debug} || $self->{debug};
     my $keep_timeline = delete $request->{keep_timeline} || $self->{keep_timeline};
@@ -158,6 +169,8 @@ sub request {
     $self->{callbacks}{$conn_id} = \%callbacks;
     $self->{connections}{$conn_id} = $conn;
 
+    _set_lifetime_timer($self, $conn_id) if $request->{lifetime_timeout};
+
     return $conn if $request->{_test}; # for testing purposes
     _set_init_state($self, $conn_id);
 
@@ -206,6 +219,11 @@ sub socket_cache {
 # Routines to manipulate connections (also user facing)
 ################################################################################
 
+sub yahc_terminal_error {
+    my $error = shift;
+    return int($error) & YAHC::Error::TERMINAL_ERROR() == YAHC::Error::TERMINAL_ERROR() ? 1 : 0;
+}
+
 sub yahc_reinit_conn {
     my ($conn, $args) = @_;
     die "YAHC: cannot reinit completed connection\n"
@@ -217,7 +235,8 @@ sub yahc_reinit_conn {
 
     my $request = $conn->{request};
     do { $request->{$_} = $args->{$_} if $args->{$_} } foreach (keys %$args);
-    $request->{_target} = _wrap_target_selection($args->{host}) if $args->{host};
+    $request->{_target}  = _wrap_host($args->{host})             if exists $args->{host};
+    $request->{_backoff} = _wrap_backoff($args->{backoff_delay}) if exists $args->{backoff_delay};
 }
 
 sub yahc_retry_conn {
@@ -335,68 +354,88 @@ sub _set_init_state {
 
     my $socket_cache = $self->{socket_cache};
     my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
-    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
 
     $conn->{response} = { status => 0 };
     $conn->{state} = YAHC::State::INITIALIZED();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if exists $conn->{debug_or_timeline};
     _call_state_callback($self, $conn, 'init_callback') if exists $conn->{has_init_callback};
 
-    my $continue = 1;
-    while ($continue) {
+    while (1) {
         _close_or_cache_socket($self, $conn, 1); # force connection close if any (likely not)
-
-        if ($conn->{attempt} > $conn->{retries}) {
-            _set_user_action_state($self, $conn_id, YAHC::Error::CONNECT_ERROR(), "retries limit reached");
-            return;
-        }
+        my $watchers = _delete_watchers_but_lifetime_timer($self, $conn_id); # implicit stop of all watchers
+        return _set_user_action_state($self, $conn_id, YAHC::Error::RETRY_LIMIT(), "retries limit reached")
+            if $conn->{attempt} > $conn->{retries};
 
         # don't move it before if statement
         my $attempt = ++$conn->{attempt};
 
-        # If later I would need timeout covering multiple attempts
-        # I can set them here under $attempt == 0 condition.
-        # Also, in _set_until_state_timer I would go to USER_ACTION_STATE
-        # instead of reinitializing connection.
-        _set_request_timer($self, $conn_id)    if $conn->{request}{request_timeout};
-        _set_connection_timer($self, $conn_id) if $conn->{request}{connect_timeout};
-        _set_drain_timer($self, $conn_id)      if $conn->{request}{drain_timeout};
+        my $request = $conn->{request};
+        _set_request_timer($self, $conn_id)    if $request->{request_timeout};
+        _set_connection_timer($self, $conn_id) if $request->{connect_timeout};
+        _set_drain_timer($self, $conn_id)      if $request->{drain_timeout};
 
-        eval {
-            my ($host, $ip, $port, $scheme) = _get_next_target($conn);
-            _register_in_timeline($conn, "Target $scheme://$host:$port ($ip:$port) chosen for attempt #$attempt") if exists $conn->{debug_or_timeline};
-
-            my $socket_cache_id; $socket_cache_id = yahc_conn_socket_cache_id($conn) if defined $socket_cache;
-            if (defined $socket_cache_id && exists $socket_cache->{$socket_cache_id}) {
-                _register_in_timeline($conn, "reuse socket '$socket_cache_id'") if $conn->{debug_or_timeline};
-                my $sock = $watchers->{_fh} = delete $socket_cache->{$socket_cache_id};
-                $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, sub {});
-                _set_write_state($self, $conn_id);
-            } else {
-                my $sock = _build_socket_and_connect($ip, $port, $conn->{request});
-                _set_wait_synack_state($self, $conn_id, $sock);
-            }
-
-            $continue = 0;
-            1;
-        } or do {
-            my $error = $@ || 'zombie error';
+        my ($host, $ip, $port, $scheme) = eval { _get_next_target($conn) };
+        if (my $error = $@) {
             _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Connection attempt failed: $error");
+            goto schedule_next_attempt;
+        }
+
+        _register_in_timeline($conn, "Target $scheme://$host:$port ($ip:$port) chosen for attempt #$attempt")
+            if exists $conn->{debug_or_timeline};
+
+        my $socket_cache_id; $socket_cache_id = yahc_conn_socket_cache_id($conn) if defined $socket_cache;
+        if (defined $socket_cache_id && exists $socket_cache->{$socket_cache_id}) {
+            _register_in_timeline($conn, "reuse socket '$socket_cache_id'") if $conn->{debug_or_timeline};
+            my $sock = $watchers->{_fh} = delete $socket_cache->{$socket_cache_id};
+            $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, sub {});
+            _set_write_state($self, $conn_id);
+            return;
+        }
+
+        my $sock = eval { _build_socket_and_connect($ip, $port) };
+        if (my $error = $@) {
+            _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Connection attempt failed: $error");
+            goto schedule_next_attempt;
         };
+
+        _set_connecting_state($self, $conn_id, $sock);
+        return;
+
+        schedule_next_attempt:
+        next unless exists $request->{_backoff};
+
+        $watchers = _delete_watchers_but_lifetime_timer($self, $conn_id); # drop all timers except lifetime
+        return _set_user_action_state($self, $conn_id, YAHC::Error::RETRY_LIMIT(), "retries limit reached")
+            if $conn->{attempt} > $conn->{retries};
+
+        my $backoff_delay = eval { $request->{_backoff}->($conn) };
+        if (my $error = $@) {
+            _set_user_action_state($self, $conn_id, YAHC::Error::CALLBACK_ERROR() | YAHC::Error::TERMINAL_ERROR(),
+                "exception in backoff callback (close connection): $error");
+            return;
+        };
+
+        $self->{loop}->now_update;
+        _register_in_timeline($conn, "setting backoff_timer to %.3fs", $backoff_delay) if exists $conn->{debug_or_timeline};
+        $watchers->{backoff_timer} = $self->{loop}->timer($backoff_delay, 0, sub {
+            _register_in_timeline($conn, "backoff timer of %.3fs expired", $backoff_delay) if exists $conn->{debug_or_timeline};
+            _set_init_state($self, $conn_id);
+        });
+        return;
     }
 }
 
-sub _set_wait_synack_state {
+sub _set_connecting_state {
     my ($self, $conn_id, $sock) = @_;
 
     my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
     my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
 
-    $conn->{state} = YAHC::State::WAIT_SYNACK();
+    $conn->{state} = YAHC::State::CONNECTING();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if exists $conn->{debug_or_timeline};
-    _call_state_callback($self, $conn, 'wait_synack_callback') if exists $conn->{has_wait_synack_callback};
+    _call_state_callback($self, $conn, 'connecting_callback') if exists $conn->{has_connecting_callback};
 
-    my $wait_synack_cb = _get_safe_wrapper($self, $conn, sub {
+    my $connecting_cb = _get_safe_wrapper($self, $conn, sub {
         my $sockopt = getsockopt($sock, SOL_SOCKET, SO_ERROR);
         if (!$sockopt) {
             _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to do getsockopt(): $!");
@@ -415,7 +454,7 @@ sub _set_wait_synack_state {
     });
 
     $watchers->{_fh} = $sock;
-    $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, $wait_synack_cb);
+    $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, $connecting_cb);
     _check_stop_condition($self, $conn) if exists $self->{stop_condition};
 }
 
@@ -467,8 +506,8 @@ sub _set_ssl_handshake_state {
     }
 
     if (!IO::Socket::SSL->start_SSL($fh, %options, SSL_startHandshake => 0)) {
-        _register_error($conn, YAHC::Error::SSL_ERROR(), "Failed to start SSL session: $IO::Socket::SSL::SSL_ERROR");
-        return _set_completed_state($self, $conn_id, 1); # XXX should be _set_init_state() ???
+        return _set_user_action_state($self, $conn_id, YAHC::Error::SSL_ERROR() | YAHC::Error::TERMINAL_ERROR(),
+            "failed to start SSL session: $IO::Socket::SSL::SSL_ERROR");
     }
 
     my $handshake_cb = _get_safe_wrapper($self, $conn, sub {
@@ -671,7 +710,7 @@ sub _set_user_action_state {
 
     $conn->{state} = YAHC::State::USER_ACTION();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if exists $conn->{debug_or_timeline};
-    _register_error($conn, $error, $strerror) if $error;
+    _register_error($conn, $error, $strerror) if $error != YAHC::Error::NO_ERROR;
 
     _close_or_cache_socket($self, $conn, $error != YAHC::Error::NO_ERROR);
     return _set_completed_state($self, $conn_id) unless exists $conn->{has_callback};
@@ -684,13 +723,19 @@ sub _set_user_action_state {
     } or do {
         my $error = $@ || 'zombie error';
         _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "Exception in user action callback (close connection): $error");
-        warn "YAHC: exception in user action callback (close connection): $error";
         $self->{state} = YAHC::State::COMPLETED();
     };
 
     $self->{loop}->now_update;
 
     my $state = $conn->{state};
+    if (yahc_terminal_error($error)) {
+        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "ignoring changed state due to terminal error")
+            unless $state == YAHC::State::USER_ACTION() || $state == YAHC::State::COMPLETED();
+        _set_completed_state($self, $conn_id, 1);
+        return
+    }
+
     _register_in_timeline($conn, "after invoking callback state is %s", _strstate($state)) if exists $conn->{debug_or_timeline};
 
     if ($state == YAHC::State::INITIALIZED()) {
@@ -723,7 +768,7 @@ sub _set_completed_state {
 }
 
 sub _build_socket_and_connect {
-    my ($ip, $port, $timeouts) = @_;
+    my ($ip, $port) = @_;
 
     my $sock;
     socket($sock, PF_INET, SOCK_STREAM, 0) or die "Failed to construct TCP socket: $!\n";
@@ -747,7 +792,7 @@ sub _get_next_target {
     # TODO STATE_RESOLVE_DNS
     ($host, $port) = ($1, $2) if !$port && $host =~ m/^(.+):([0-9]+)$/o;
     $ip = $host if !$ip && $host =~ m/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/o;
-    $ip ||= inet_ntoa(gethostbyname($host) or die "Failed to resolve '$host': $!\n");
+    $ip ||= inet_ntoa(gethostbyname($host) or die "Failed to resolve '$host': $!");
     $scheme ||= $conn->{request}{scheme} || 'http';
     $port   ||= $conn->{request}{port} || ($scheme eq 'https' ? 443 : 80);
 
@@ -832,7 +877,29 @@ sub _set_until_state_timer {
 
     $self->{loop}->now_update;
     my $w = $watchers->{$timer_name} = $self->{loop}->timer_ns($timeout, 0, $timer_cb);
-    $w->priority(-2); # set lowest priority
+    $w->priority(2); # set highest priority
+    $w->start;
+}
+
+sub _set_lifetime_timer {
+    my ($self, $conn_id) = @_;
+
+    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
+    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
+
+    delete $watchers->{lifetime_timer}; # implicit stop
+    my $timeout = $conn->{request}{lifetime_timeout};
+    return unless $timeout;
+
+    _register_in_timeline($conn, "setting lifetime timer to %.3fs", $timeout) if exists $conn->{debug_or_timeline};
+
+    $self->{loop}->now_update;
+    my $w = $watchers->{lifetime_timer} = $self->{loop}->timer_ns($timeout, 0, sub {
+        _set_user_action_state($self, $conn_id, YAHC::Error::LIFETIME_TIMEOUT() | YAHC::Error::TERMINAL_ERROR(),
+            sprintf("lifetime timer of %.3fs expired", $timeout)) if $conn->{state} < YAHC::State::COMPLETED();
+    });
+
+    $w->priority(2); # set highest priority
     $w->start;
 }
 
@@ -891,18 +958,41 @@ sub _parse_http_headers {
 # Helpers
 ################################################################################
 
-sub _wrap_target_selection {
-    my $host = shift;
-    my $ref = ref($host);
+sub _delete_watchers_but_lifetime_timer {
+    my ($self, $conn_id) = @_;
 
-    return sub { $host } if $ref eq '';
-    return $host         if $ref eq 'CODE';
+    my $watchers = $self->{watchers}{$conn_id};
+    if (defined $watchers && (my $w = $watchers->{lifetime_timer})) {
+        return $self->{watchers}{$conn_id} = { lifetime_timer => $w };
+    }
+
+    return $self->{watchers}{$conn_id} = {};
+}
+
+sub _wrap_host {
+    my ($value) = @_;
+    my $ref = ref($value);
+
+    return sub { $value } if $ref eq '';
+    return $value         if $ref eq 'CODE';
 
     my $idx = 0;
-    return sub { $host->[$idx++ % @$host]; } if $ref eq 'ARRAY';
+    return sub { $value->[$idx++ % @$value]; }
+        if $ref eq 'ARRAY' && @$value > 0;
+
     die "YAHC: unsupported host format\n";
 }
 
+sub _wrap_backoff {
+    my ($value) = @_;
+    my $ref = ref($value);
+
+    return sub { $value } if $ref eq '';
+    return $value         if $ref eq 'CODE';
+
+    die "YAHC: unsupported backoff format\n";
+}
+
 sub _call_state_callback {
     my ($self, $conn, $cb_name) = @_;
     my $cb = $self->{callbacks}{$conn->{id}}{$cb_name};
@@ -913,8 +1003,7 @@ sub _call_state_callback {
         1;
     } or do {
         my $error = $@ || 'zombie error';
-        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "Exception in state callback (ignore error): $error");
-        warn "YAHC: exception in state callback (ignore error): $error";
+        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "exception in state callback (ignore error): $error");
     };
 
     # $self->{loop}->now_update; # XXX expect state callbacks to be small
@@ -927,9 +1016,8 @@ sub _get_safe_wrapper {
         1;
     } or do {
         my $error = $@ || 'zombie error';
-        _register_error($conn, YAHC::Error::INTERNAL_ERROR(), "Exception callback: $error");
-        warn "YAHC: exception in callback: $error";
-        _set_completed_state($self, $conn->{id}, 1);
+        _set_user_action_state($self, $conn->{id}, YAHC::Error::INTERNAL_ERROR() | YAHC::Error::TERMINAL_ERROR(),
+            "exception in internal callback: $error");
     }};
 }
 
@@ -943,7 +1031,7 @@ sub _register_in_timeline {
 sub _register_error {
     my ($conn, $error, $format, @arguments) = @_;
     my $strerror = sprintf("$format", @arguments);
-    _register_in_timeline($conn, "error=$strerror ($error)") if exists $conn->{debug_or_timeline};
+    _register_in_timeline($conn, "strerror='$strerror' error=$error") if exists $conn->{debug_or_timeline};
     push @{ $conn->{errors} ||= [] }, [ $error, $strerror, [ @{ $conn->{selected_target} } ], Time::HiRes::time ];
 }
 
@@ -951,7 +1039,7 @@ sub _strstate {
     my $state = shift;
     return 'STATE_INIT'         if $state eq YAHC::State::INITIALIZED();
     return 'STATE_RESOLVE_DNS'  if $state eq YAHC::State::RESOLVE_DNS();
-    return 'STATE_WAIT_SYNACK'  if $state eq YAHC::State::WAIT_SYNACK();
+    return 'STATE_CONNECTING'   if $state eq YAHC::State::CONNECTING();
     return 'STATE_CONNECTED'    if $state eq YAHC::State::CONNECTED();
     return 'STATE_WRITING'      if $state eq YAHC::State::WRITING();
     return 'STATE_READING'      if $state eq YAHC::State::READING();
@@ -1027,7 +1115,7 @@ Each YAHC connection goes through following list of states in its lifetime:
               v   +-----------------+   ^
               v           |             ^
               v   +-------v---------+   ^
-              +<<-+   WAIT SYNACK   +->>+
+              +<<-+    CONNECTING   +->>+
               v   +-----------------+   ^
               v           |             ^
      Path in  v   +-------v---------+   ^  Retry
@@ -1061,7 +1149,7 @@ In normal situation a connection after being initialized goes through state:
 
 - RESOLVE DNS
 
-- WAIT SYNACK - wait finishing of handshake
+- CONNECTING - wait finishing of handshake
 
 - CONNECTED
 
@@ -1082,9 +1170,6 @@ In case of IO error during normal execution YAHC retries connection
 C<retries> times. In practise this means that connection goes back to
 INITIALIZED state.
 
-It's possible for a connection to go directly to COMPLETED state in case
-of internal error.
-
 =item 3) Failure path (left line).
 
 If all retry attempts did not succeeded a connection goes to state 'USER
@@ -1104,10 +1189,16 @@ to give a change to user to interupt the workflow.
 
 =item * HTTP response received. Note that non-200 responses are NOT treated as error. 
 
-=item * unsupported HTTP response is received (such as response without Content-Length header
+=item * unsupported HTTP response is received (such as response without Content-Length header)
 
 =item * retries limit reached
 
+=item * lifetime timeout has expired
+
+=item * provided callback has thrown exception
+
+=item * internal error has occured
+
 =back
 
 When a connection enters this state C<callback> CodeRef is called:
@@ -1122,7 +1213,7 @@ When a connection enters this state C<callback> CodeRef is called:
             ) = @_;
 
             # Note that fields in $conn->{response} are not set 
-            # if $error != # YAHC::Error::NO_ERROR()
+            # if $error != YAHC::Error::NO_ERROR()
 
             # HTTP response is stored in $conn->{response}.
             # It can be also accessed via yahc_conn_response().
@@ -1145,6 +1236,11 @@ C<yahc_reinit_conn> may be called to give the request more chances to complete
 successfully (for example by following redirects or providing new target
 hosts).
 
+In some cases connection cannot be retried anymore and callback is
+called for information purposes only. This case can be distinguished by
+C<$error> having YAHC::Error::TERMINAL_ERROR() bit set. One can use
+C<yahc_terminal_error> helper to detect such case.
+
 Note that C<callback> should NOT throw exception. If so the connection will be
 imidiately closed.
 
@@ -1243,15 +1339,22 @@ wise to set C<account_for_signals>.
     connect_timeout        => undef,
     request_timeout        => undef,
     drain_timeout          => undef,
+    lifetime_timeout       => undef,
+
+    # burst control
+    backoff_delay          => undef,
 
     # callbacks
     init_callback          => undef,
-    wait_synack_callback   => undef,
+    connecting_callback    => undef,
     connected_callback     => undef,
     writing_callback       => undef,
     reading_callback       => undef,
     callback               => undef,
 
+    # SSL options
+    ssl_options            => {},
+
 Notice how YAHC does not take a full URI string as input, you have to
 specify the individual parts of the URL. Users who need to parse an
 existing URI string to produce a request should use the L<URI> module
@@ -1317,24 +1420,70 @@ C<host> parameter can accept one of following values:
 The value of C<connect_timeout>, C<request_timeout> and C<drain_timeout> is in
 floating point seconds, and is used as the time limit for connecting to the
 host (reaching CONNECTED state), full request time (reaching COMPLETED state)
-and sending request to remote site (reaching READING state) respectively. The
-default value for all is C<undef>, meaning no timeout limit. If you don't
-supply these timeouts and the host really is unreachable or slow, we'll reach
-the TCP timeout limit before returning some other error to you.
+and sending request to remote site (reaching READING state) respectively.
+
+C<lifetime_timeout> has special purpose. Its task is to provide upper bound
+timeout for a request lifetime. In other words, if a request comes with
+multiple retries C<connect_timeout>, C<request_timeout> and C<drain_timeout>
+are per attempt. C<lifetime_timeout> covers all attempts. If by the time
+C<lifetime_timeout> expires a connection is not in COMPLETED state a error is
+generated. Note that after this error the connection cannot be retried anymore.
+So, it's forced to go to COMPLETED state.
+
+The default value for all is C<undef>, meaning no timeout limit.
+
+=head3 backoff_delay
+
+C<backoff_delay> can be used to introduce delay between retries. This is a
+great way to avoid load spikes on server side. Following example creates new
+request which would be retried twice doing three attempts in total. Second and
+third attempts will be delay by one second each.
+
+    $yach->request({
+        host          => "example.com",
+        retries       => 2,
+        backoff_delay => 1,
+    });
+
+
+C<backoff_delay> can be set in two ways:
+
+=over 4
+
+    1) floating point seconds - define constant delay between retires.
+
+    2) CodeRef. The subroutine is invoked on each retry and should return
+    floating point seconds. This option is useful for having exponentially
+    growing delay or, for instance, jitted delays.
+
+=back
+
+The default value is C<undef>, meaning no delay.
 
 =head3 callbacks
 
-The value of C<init_callback>, C<wait_synack_callback>, C<connected_callback>,
+The value of C<init_callback>, C<connecting_callback>, C<connected_callback>,
 C<writing_callback>, C<reading_callback> is CodeRef to a subroutine which is
 called upon reaching corresponding state. Any exception thrown in the
-subroutine moves connection to COMPLETED state effectively terminating any
-ongoing IO.
+subroutine will be ignored.
 
 The value of C<callback> defines main request callback which is called when a
 connection enters 'USER ACTION' state (see 'USER ACTION' state above).
 
 Also see L<LIMITATIONS>
 
+=head3 ssl_options
+
+Performing HTTPS request the value of ssl_options extended by two parameters
+set to current hostname:
+
+        SSL_verifycn_name => $hostname,
+        IO::Socket::SSL->can_client_sni ? ( SSL_hostname => $hostname ) : (),
+
+Apart of this changes, the value is directly passed to
+C<IO::Socket::SSL::start_SSL()>. For more details refer to IO::Socket::SSL
+documentation L<https://metacpan.org/pod/IO::Socket::SSL>.
+
 =head2 drop
 
 Given connection HashRef or conn_id move connection to COMPLETED state (avoiding
@@ -1358,6 +1507,10 @@ call:
 
     $yahc->run(YAHC::State::COMPLETED(), $conn_id);
 
+or simply:
+
+    $yahc->run();
+
 Leaving list of connection empty makes YAHC waiting for all connection reaching
 needed until_state.
 
@@ -1402,9 +1555,11 @@ One of use cases of C<yahc_reinit_conn>, for example, is to handle redirects:
     $yahc->request({
         host => 'domain_which_returns_301.com',
         callback => sub {
+            ...
             my $conn = $_[0];
             yahc_reinit_conn($conn, { host => 'www.newtarget.com' })
                 if $_[0]->{response}{status} == 301;
+            ...
         }
     });
 
@@ -1418,6 +1573,29 @@ connection is in 'USER ACTION' state.
 Retries given connection. C<yahc_retry_conn> should be called only if
 C<yahc_conn_attempts_left> returns positive value. Otherwise, it exits silently.
 
+Intended usege is to retry transient failures or try different host:
+
+    use YAHC qw/
+        yahc_retry_conn
+        yahc_conn_attempts_left
+    /;
+
+    my ($yahc, $yahc_storage) = YAHC->new();
+    $yahc->request({
+        retries => 2,
+        host => [ 'host1', 'host2' ],
+        callback => sub {
+            ...
+            my $conn = $_[0];
+            if ($_[0]->{response}{status} == 503 && yahc_conn_attempts_left($conn)) {
+                yahc_retry_conn($conn);
+            }
+            ...
+        }
+    });
+
+    $yahc->run;
+
 C<yahc_conn_attempts_left> is meant to be called inside C<callback> similarly
 to C<yahc_reinit_conn>.
 
@@ -1458,6 +1636,11 @@ items:
 
 Return last error appeared in connection. See C<yahc_conn_errors>.
 
+=head2 yahc_terminal_error
+
+Given a error return 1 if the error has YAHC::Error::TERMINAL_ERROR() bit set.
+Otherwise return 0.
+
 =head2 yahc_conn_timeline
 
 Return timeline of given connection. See more about timeline in description of
